# Copyright 2021 Olivier 'reivilibre'
#  This file is part of https://bics.ga/reivilibre/matrix_queue_proc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import os
import subprocess
import sys
from asyncio import Semaphore
from getpass import getpass
from pathlib import Path
from typing import Union, List

from nio import (
    AsyncClient,
    AsyncClientConfig,
    LoginResponse,
    LoginError,
    SyncResponse,
    RoomGetStateError,
    RoomGetStateResponse,
    RoomGetStateEventError,
    RoomGetStateEventResponse,
    RoomContextError,
    RoomContextResponse,
    BadEvent,
    Event,
    UnknownBadEvent,
    MegolmEvent,
    EncryptionError,
    RoomMessageText,
    RoomMessageNotice,
    MessageDirection,
    RoomMessagesError,
    RoomMessagesResponse,
)
from nio.client.async_client import ResponseCb
from nio.store import SqliteStore


class QueueProcBot:
    def __init__(self, data_path: Path):
        self.client: AsyncClient = None  # type: ignore
        self.data_path = data_path
        self._sync_forever_task = None

        data_path.mkdir(exist_ok=True)

    async def connect(self):
        cconfig = AsyncClientConfig(
            # store=SqliteStore
        )

        login_info_path = Path(self.data_path, "login_info.json")
        reusing_login = login_info_path.exists()

        user_id = None
        username = None
        password = None
        access_token = None
        device_id = None

        if reusing_login:
            print(f"Reusing connection from {login_info_path}.", file=sys.stderr)
            with open(login_info_path, "r") as fin:
                login_info = json.load(fin)
            homeserver = login_info["homeserver"]
            user_id = login_info["user_id"]
            access_token = login_info["access_token"]
            device_id = login_info["device_id"]
        else:
            print("New connection.")
            homeserver = input("Enter homeserver such as https://matrix.org: ")
            username = input("Enter username: ")
            password = getpass("Enter Matrix password")

        store_path = Path(self.data_path, "nio_store")
        store_path.mkdir(exist_ok=True)
        self.client = AsyncClient(
            homeserver, user_id or username, store_path=store_path, config=cconfig
        )

        if reusing_login:
            self.client.restore_login(user_id, device_id, access_token)
            return True
        else:
            lrp = await self.client.login(
                password, device_name="Matrix Queue Processor"
            )
            if isinstance(lrp, LoginResponse):
                login_info = {
                    "homeserver": homeserver,
                    "user_id": lrp.user_id,
                    "device_id": lrp.device_id,
                    "access_token": lrp.access_token,
                }
                with open(login_info_path, "w") as fout:
                    print("Saving login information.", file=sys.stderr)
                    json.dump(login_info, fout)
                return True
            elif isinstance(lrp, LoginError):
                print(f"Error logging in\n\t{lrp}", file=sys.stderr)
                return False

    async def startup(self):
        monitor = Semaphore(value=0)
        resp_cb = None

        async def good_to_go(_resp):
            print("Synced!")
            monitor.release()
            self.client.response_callbacks.remove(resp_cb)

        resp_cb = ResponseCb(good_to_go, SyncResponse)

        self.client.response_callbacks.append(resp_cb)
        self._sync_forever_task = asyncio.create_task(self.client.sync_forever())

        await monitor.acquire()

    async def process_events(
        self,
        events: List[Union[Event, BadEvent, UnknownBadEvent]],
        program: str,
        room_id: str,
        delete_on_success: bool,
    ):
        for event in events:
            # print(event.event_id, event.__class__.__name__)
            if isinstance(event, MegolmEvent):
                print("!!! Megolm Event :(", event.event_id)
                try:
                    await asyncio.sleep(2)
                    print(self.client.decrypt_event(event))
                except EncryptionError:
                    print("failed to decrypt")
            elif isinstance(event, RoomMessageText):
                text = event.body
                print(f"Task: {text}")

                exit_code = await asyncio.get_event_loop().run_in_executor(
                    None, subprocess.call, [program, text]
                )

                if exit_code == 0:
                    if delete_on_success:
                        await self.client.room_redact(room_id, event.event_id)
                    else:
                        await self.client.room_send(
                            room_id,
                            "m.reaction",
                            {
                                "m.relates_to": {
                                    "rel_type": "m.annotation",
                                    "event_id": event.event_id,
                                    "key": "✅",
                                }
                            },
                            ignore_unverified_devices=True,
                        )
                else:
                    await self.client.room_send(
                        room_id,
                        "m.room.message",
                        {
                            "m.relates_to": {
                                "m.in_reply_to": {"event_id": event.event_id}
                            },
                            "msgtype": "m.notice",
                            "body": f"Failed with exit code {exit_code}.",
                        },
                        ignore_unverified_devices=True,
                    )

                # update our marker
                await self.client.room_read_markers(
                    room_id, event.event_id, event.event_id
                )

                # cooldown.
                await asyncio.sleep(5)

    async def catch_up(self, room_id: str, program: str, delete_on_success: bool):
        room = self.client.rooms[room_id]

        start_at = room.fully_read_marker

        if start_at is None:
            print("No fully read marker")
            room_creation_event = await self.client.room_get_state_event(
                room_id, "m.room.create"
            )
            if isinstance(room_creation_event, RoomGetStateEventError):
                raise RuntimeError(str(room_creation_event))
            assert isinstance(room_creation_event, RoomGetStateEventResponse)

            start_at = room_creation_event.content["event_id"]

        print("Starting at: ", start_at)

        context = await self.client.room_context(room_id, start_at)
        if isinstance(context, RoomContextError):
            raise RuntimeError(str(context))
        assert isinstance(context, RoomContextResponse)

        await self.process_events(
            context.events_after, program, room_id, delete_on_success
        )

        next_batch = context.end

        while True:
            messages = await self.client.room_messages(
                room_id,
                next_batch,
                direction=MessageDirection.front,
            )
            if isinstance(messages, RoomMessagesError):
                raise RuntimeError(str(messages))
            assert isinstance(messages, RoomMessagesResponse)

            next_batch = messages.end

            if len(messages.chunk) == 0:
                break

            await self.process_events(
                messages.chunk, program, room_id, delete_on_success
            )


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print("Usage: mxprocqueue.py <room_id> <processing script>")
        print("Set MXQP_REDACT=1 to redact messages on success.")
        sys.exit(127)
    room_id = sys.argv[1]
    command = sys.argv[2]

    bot = QueueProcBot(Path("./queueprocdata"))  # TODO configurability

    async def main():
        print("Logging in...")
        await bot.connect()
        print("Syncing...")
        await bot.startup()
        print("Catching up...")
        delete_on_success = os.environ.get("MXQP_REDACT") == "1"
        print("DoS", delete_on_success)
        await bot.catch_up(room_id, command, delete_on_success)
        print("Done?")

    asyncio.get_event_loop().run_until_complete(main())
